# -*- coding: utf-8 -*-
"""Universal converter for time units. Usage:

.. code:: python

    from time_converter import Time
    Time(1000, 'sol').to('datetime')

Currently supported units:

- ``'datetime', 'dt'``: Python datetime objects
- ``'sol'``: MSL sol numbers (float)
- ``'sclk'``: MSL spacecraft clock (float)
- ``'doy'``: Tuple of (year, doy) (year: int, doy: float)
- ``'decimalyear', 'dy'``: Decimal year (float, e.g. 2018.086)
- ``'posix'``: POSIX timestamp (seconds since Jan 1 1970, 00:00 UTC)

Providing the unit name is optional for converting from datetime as it can be inferred from the data type, e.g.

.. code:: python

    from time_converter import Time
    import datetime as dt
    Time(dt.datetime(2018, 1, 1)).to('sol')

Author: Johan von Forstner <forstner@physik.uni-kiel.de>, January 2018
"""

import datetime as dt
from abc import ABCMeta, abstractmethod

try:
    import pandas as pd
except:
    pd = None

import numpy as np
from six import with_metaclass


class Converter(with_metaclass(ABCMeta)):
    """
    Abstract class that defines a converter for converting between some time unit <-> Python's datetime class
    """

    @abstractmethod
    def supports(self, unit=None, datatype=None):
        # type: (str, type) -> bool
        """
        Determines if this converter can convert to or from the given data type

        :param unit: string description of the unit supplied by the user, can be None when not supplied
        :param datatype: class of the data that is supplied (e.g. float, datetime, ...) to convert from, or None
        """
        pass

    @abstractmethod
    def convert_from_datetime(self, datetime):
        # type: (dt.datetime) -> object
        """
        Converts the given datetime value to the unit supported by this converter

        :param datetime: datetime value
        """
        pass

    @abstractmethod
    def convert_to_datetime(self, value):
        # type: (object) -> dt.datetime
        """
        Converts the given value from the unit supported by this converter to a datetime value

        :param value:
        """
        pass


from time_converter.converters.earth import *
from time_converter.converters.msl import *
from time_converter.converters.change4 import *


class Time:
    converters = [clazz() for clazz in Converter.__subclasses__()]  # instantiate one converter of every kind

    def __init__(self, value, unit=None):
        if self._is_list(value):
            if len(value) == 0:
                self.dt = []
            else:
                converter = self._get_converter(unit, type(value[0]))
                self.dt = [converter.convert_to_datetime(val) for val in value]
            self.original_value = value
        else:
            converter = self._get_converter(unit, type(value))
            self.dt = converter.convert_to_datetime(value)

    def to(self, unit):
        converter = self._get_converter(unit)
        if self._is_list(self.dt):
            arr = np.array([converter.convert_from_datetime(d) for d in self.dt])
            if pd is not None and type(self.original_value) == pd.core.series.Series:
                return pd.Series(arr, index=self.original_value.index)
            else:
                return arr
        else:
            return converter.convert_from_datetime(self.dt)

    def _get_converter(self, unit, datatype=None):
        for converter in self.converters:
            if converter.supports(unit, datatype):
                return converter

        raise ValueError('unknown unit {} or unsupported type {}'.format(unit, datatype))

    def _is_list(self, value):
        return type(value) in [np.ndarray, list] or pd is not None and type(value) == pd.core.series.Series



import spiceypy as spice
import datetime as dt
import math
import pandas as pd

SPICE_KERNELS_PATH = '/home/asterix/forstner/git/spiceminer/data'

spice.furnsh('{}/base/spk/planets/de430.bsp'.format(SPICE_KERNELS_PATH))
spice.furnsh('{}/base/lsk/naif0011.tls'.format(SPICE_KERNELS_PATH))
spice.furnsh('{}/base/pck/pck00010.tpc'.format(SPICE_KERNELS_PATH))

CHANGE4_LONGITUDE = 177.6
# source: http://www.cnsa.gov.cn/english/n6465652/n6465653/c6805049/content.html

moon = spice.bodn2c('MOON')


def get_localtime(time, body, lon):
    h, m, s, _, _ = spice.et2lst(spice.datetime2et(time), body, lon, 'PLANETOCENTRIC')
    return dt.time(h, m, s)


def get_change4_localtime(time):
    return get_localtime(time, moon, math.radians(177.6))

if __name__ == '__main__':
    index = pd.date_range(freq='30min', start=dt.datetime(2018, 12, 22, 16), end=dt.datetime(2022, 1, 1))
    localtimes = pd.Series([get_change4_localtime(val) for val in index], index=index)

    datetimes = pd.Series([dt.datetime.combine(dt.date(2019, 1, 1), val) for val in localtimes], index=index)
    new_lunar_day = datetimes.diff() < dt.timedelta(seconds=0)
    lunar_days = pd.Series([new_lunar_day[:val].sum() for val in index], index=index)

    df = pd.DataFrame({'LD': lunar_days, 'time': localtimes}, index=index)
    df.to_csv('converters/change4/change4_localtime.dat')



# -*- coding: utf-8 -*-
"""Replicates functionality of NASA NAIF's chronos tool to convert MSL SCLK time to UTC

It uses the msl.tsc file, which was downloaded from
https://naif.jpl.nasa.gov/pub/naif/MSL/kernels/sclk/msl.tsc
and might be updated regularly

Author: Johan von Forstner <forstner@physik.uni-kiel.de>, January 2018
"""

import datetime as dt
import os

import numpy as np

sclk_data = None


def _load_sclk_data(filename=os.path.join(os.path.dirname(__file__), 'msl.tsc')):
    """
    Loads the data from the "Source SCLKvSCET File" section of the msl.tsc file

    :rtype: tuple
    :type filename: str
    :param filename: the file name (default: msl.tsc)
    :return: data from the four columns: sclk, utc, dut, sclkrate
    """
    global sclk_data
    if sclk_data is None:
        file = open(filename)
        correct_section = False
        skip_header = None
        max_rows = None
        for i, line in enumerate(file):
            if skip_header is not None and line.lstrip().startswith('CCSD'):
                max_rows = i - skip_header
                break
            elif correct_section and line.lstrip().startswith('*'):
                skip_header = i + 1
            elif line.startswith('Source SCLKvSCET File'):
                correct_section = True
        data = np.genfromtxt(filename, dtype=(float, dt.datetime, float, float), skip_header=skip_header, max_rows=max_rows,
                             unpack=True,
                             converters={
                                 0: np.float,
                                 1: lambda x: dt.datetime.strptime(x.decode('ascii'), "%Y-%jT%H:%M:%S.%f"),
                                 2: np.float,
                                 3: np.float
                             })
        sclk, utc, dut, sclkrate = data['f0'], data['f1'], data['f2'], data['f3']
        sclk_data = sclk, utc, dut, sclkrate
    return sclk_data


def sclk_to_dt(sclk):
    """
    Accurately converts spacecraft clock (SCLK) to UTC datetime using the chronos data

    :rtype: dt.datetime
    :type sclk: Real
    :param sclk: SCLK to convert
    :return: datetime as UTC
    """
    sclks, dts, dut, sclkrate = _load_sclk_data()
    # find index of last position with sclk smaller or equal to the given one
    index = np.searchsorted(sclks, sclk, side='right') - 1
    # extrapolate using sclkrate given in table
    seconds = (sclk - sclks[index]) * sclkrate[index]
    return dts[index] + dt.timedelta(seconds=seconds)


def dt_to_sclk(dt):
    """
    Accurately converts UTC datetime to spacecraft clock (SCLK) using the chronos data

    :type: dt: dt.datetime
    :rtype float
    :param dt: UTC datetime to convert
    :return: corresponding SCLK value
    """
    sclks, dts, dut, sclkrate = _load_sclk_data()
    # find index of last position with sclk smaller or equal to the given one
    index = np.searchsorted(dts, dt, side='right') - 1
    # extrapolate using sclkrate given in table
    return sclks[index] + (dt - dts[index]).total_seconds() / sclkrate[index]


"""
This module includes tests for the time_Timeer module
"""
import pytest

from time_converter import Time
from pytest import approx, raises
import datetime as dt
import numpy as np


def test_consistency():
    assert Time(1000, 'sol').to('sol') == approx(1000)
    assert Time(dt.datetime(2018, 1, 1)).to('dt') == dt.datetime(2018, 1, 1)
    assert Time(6e8, 'sclk').to('sclk') == approx(6e8)
    assert Time((2018, 2.5), 'doy').to('doy') == (2018, 2.5)
    assert Time(2018.5, 'decimalyear').to('decimalyear') == approx(2018.5)
    assert Time(1517742867, 'posix').to('posix') == 1517742867

    ce4lst = Time((1, dt.time(8, 30)), 'ce4lst').to('ce4lst')
    assert ce4lst[0] == 1
    assert abs(dt.datetime.combine(dt.date.today(), ce4lst[1]) - dt.datetime.combine(dt.date.today(), dt.time(8, 30))) < dt.timedelta(minutes=1)


def test_values():
    time = Time(dt.datetime(2018, 1, 15))
    assert time.to('sol') == approx(1935.221950230014)
    assert time.to('sclk') == approx(569244654.78994048)
    assert time.to('doy') == (2018, approx(15))
    assert time.to('decimalyear') == approx(2018.0383561643835)
    assert time.to('posix') == 1515974400
    with pytest.raises(ValueError):
        time.to('ce4lst')

    time = Time(dt.datetime(2019, 3, 20))
    assert time.to('ce4lst') == (3, dt.time(22, 49, 2))


def test_values_reverse():
    try:
        import spiceypy as spice
    except ImportError:
        spice = None

    if spice is not None:
        assert Time((1, dt.time(0, 0)), 'ce4lst').to('dt') == dt.datetime(2018, 12, 22, 16, 25, 51, 600000)
        assert Time((1, dt.time(23, 59, 59)), 'ce4lst').to('dt') == dt.datetime(2019, 1, 21, 7, 3, 44, 200000)
        assert Time((10, dt.time(23, 59, 59)), 'ce4lst').to('dt') == dt.datetime(2019, 10, 13, 20, 50, 37, 900000)
    else:
        assert Time((1, dt.time(0, 0)), 'ce4lst').to('dt') == dt.datetime(2018, 12, 22, 16, 30)
        assert Time((1, dt.time(23, 59, 59)), 'ce4lst').to('dt') == dt.datetime(2019, 1, 21, 7, 30)


def test_empty():
    time = Time([])
    assert np.array_equal(time.to('sol'), [])


def test_error():
    with raises(ValueError):
        Time(1, 'foo')


def test_pandas():
    import pandas as pd
    series = pd.Series([1, 2])
    converted = Time(series, 'sol').to('sol')
    assert (converted == series).all()
    assert type(converted) == pd.core.series.Series



from time_converter import Converter

from . import chronos
import datetime as dt
from numbers import Real

SOL = 88775.24409  # seconds per sol
LANDING = dt.datetime(2012, 8, 5, 13, 49, 59)


class SolConverter(Converter):
    def supports(self, unit=None, datatype=None):
        return (datatype is None or issubclass(datatype, Real)) and (unit == 'sol')

    def convert_from_datetime(self, datetime):
        return (datetime - LANDING).total_seconds() / SOL

    def convert_to_datetime(self, value):
        return LANDING + dt.timedelta(seconds=value * SOL)


class SclkConverter(Converter):
    def supports(self, unit=None, datatype=None):
        return (datatype is None or issubclass(datatype, Real)) and (unit == 'sclk')

    def convert_from_datetime(self, datetime):
        return chronos.dt_to_sclk(datetime)

    def convert_to_datetime(self, value):
        return chronos.sclk_to_dt(value)



import os

from time_converter import Converter
import pandas as pd
import datetime as dt
import warnings

from time_converter.converters.spice_utilities import setup_spice, datetime_to_solarday_number, datetime_to_lst, \
    lst_to_datetime

try:
    import spiceypy as spice
except ImportError:
    spice = None

FIRST_LUNARDAY = dt.datetime(2018, 12, 22, 16, 25, 51, 600000)

CHANGE4_LONGITUDE = 177.6
# source: http://www.cnsa.gov.cn/english/n6465652/n6465653/c6805049/content.html


class Change4LocalTimeConverter(Converter):
    _data = None
    _day_cache = []

    def _load_data(self):
        warnings.warn('Using precomputed data for Chang\'E 4 local time calculation.\n'
                      'Run:\n'
                      'pip3 install --user time_converter[spice]\n'
                      'to use spiceypy for better accuracy and performance.')
        if self._data is not None:
            return self._data
        else:
            filename = os.path.join(os.path.dirname(__file__), 'change4_localtime.dat')
            data = pd.read_csv(filename, index_col=0, parse_dates=[0, 2])
            data['time'] = pd.Series([val.time() for val in data['time']], index=data.index)
            self._data = data
            return data

    def supports(self, unit=None, datatype=None):
        return (datatype is None or issubclass(datatype, tuple)) and (unit == 'ce4lst')

    def convert_from_datetime(self, datetime):
        if spice is not None:
            setup_spice()

            if datetime < FIRST_LUNARDAY:
                raise ValueError('unsupported date: {}'.format(datetime))

            day = datetime_to_solarday_number(datetime, FIRST_LUNARDAY, 'MOON', CHANGE4_LONGITUDE, self._day_cache)
            time = datetime_to_lst(datetime, 'MOON', CHANGE4_LONGITUDE)
            return day, time
        else:
            data = self._load_data()
            value = pd.Timestamp(datetime)

            if value > data.index.max() or value < data.index.min():
                raise ValueError('unsupported date: {}'.format(datetime))

            row = data.asof(value)
            return row['LD'], row['time']

    def convert_to_datetime(self, value):
        if spice is not None:
            setup_spice()
            if value[0] < 1:
                raise ValueError('unsupported lunar day: {}'.format(value[0]))

            return lst_to_datetime(value[0], value[1], FIRST_LUNARDAY, 'MOON', CHANGE4_LONGITUDE, self._day_cache)
        else:
            data = self._load_data()
            if value[0] < 1 or value[0] >= data['LD'].max():
                raise ValueError('unsupported lunar day: {}'.format(value[0]))

            masked_data = data[(data['LD'] == value[0]) & (data['time'] >= value[1])]
            if len(masked_data) == 0:
                masked_data = data[data['LD'] > value[0]]
            return pd.Timestamp(masked_data.index.values[0]).to_pydatetime()


from time_converter import Converter
import datetime as dt
from numbers import Real
import calendar


class DatetimeConverter(Converter):
    def supports(self, unit=None, datatype=None):
        return (datatype is None or issubclass(datatype, dt.datetime)) and\
                   (unit is None or unit == 'dt' or unit == 'datetime')

    def convert_from_datetime(self, datetime):
        return datetime

    def convert_to_datetime(self, value):
        return value


class DoyConverter(Converter):
    def supports(self, unit=None, datatype=None):
        return (datatype is None or issubclass(datatype, tuple)) and (unit == 'doy')

    def convert_from_datetime(self, datetime):
        ttuple = datetime.utctimetuple()
        return ttuple.tm_year, \
               ttuple.tm_yday + ttuple.tm_hour / 24. + ttuple.tm_min / 24. / 60. + ttuple.tm_sec / 24. / 60. / 60.

    def convert_to_datetime(self, value):
        yy, day = value
        return dt.datetime(int(yy) - 1, 12, 31) + dt.timedelta(days=day)


class DecimalYearConverter(Converter):
    def supports(self, unit=None, datatype=None):
        return (datatype is None or issubclass(datatype, Real)) and (unit == 'dy' or unit == 'decimalyear')

    def convert_from_datetime(self, datetime):
        year_part = datetime - dt.datetime(year=datetime.year, month=1, day=1)
        year_length = dt.datetime(year=datetime.year + 1, month=1, day=1) - dt.datetime(year=datetime.year, month=1,
                                                                                        day=1)
        return datetime.year + year_part.total_seconds() / year_length.total_seconds()

    def convert_to_datetime(self, value):
        year = int(value)
        year_length = dt.datetime(year=year + 1, month=1, day=1) - dt.datetime(year=year, month=1, day=1)
        return dt.datetime(year, 1, 1) + dt.timedelta(seconds=year_length.total_seconds() * (value - year))


class PosixConverter(Converter):
    def supports(self, unit=None, datatype=None):
        return (datatype is None or issubclass(datatype, Real)) and (unit == 'posix')

    def convert_from_datetime(self, datetime):
        return calendar.timegm(datetime.utctimetuple())

    def convert_to_datetime(self, value):
        return dt.datetime.utcfromtimestamp(value)



"""Setup module, used for packaging time_converter to make it installable with pip
based on: https://github.com/pypa/sampleproject/blob/master/setup.py
"""

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='time_converter',

    version='1.2.2',

    description='A Python class that allows for convenient conversion between different date and time formats/units',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitlab.physik.uni-kiel.de/ET/time_converter',
    author='Johan von Forstner',
    author_email='forstner@physik.uni-kiel.de',

    packages=find_packages(),
    install_requires=['numpy', 'python-dateutil', 'tqdm', 'requests', 'pandas'],
    extras_require={
        'dev': ['pytest', 'pytest-cov', 'numpydoc', 'Sphinx'],
        'spice': ['spiceypy']
    },
    package_data={
        'time_converter.converters': [
            'msl/msl.tsc',
            'change4/change4_localtime.dat'
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import math
import datetime as dt
import dateutil.parser
import numpy as np
import os
from tqdm import tqdm
import requests

try:
    import spiceypy as spice
except ImportError:
    spice = None


def download_file(url, filename):
    r = requests.get(url, stream=True)
    size = int(r.headers.get('content-length', 0))
    block_size = 8 * 1024
    with open(filename, 'wb') as file:
        pbar = tqdm(total=size, unit='B', unit_scale=True,
                    desc='downloading {}'.format(url.split('/')[-1]))
        for chunk in r.iter_content(block_size):
            file.write(chunk)
            pbar.update(len(chunk))
        pbar.close()


def setup_spice():
    if spice.ktotal('all') > 0: return

    SPICE_KERNELS_PATH = os.path.join(os.path.expanduser('~'), '.time_converter', 'spicefiles')
    os.makedirs(SPICE_KERNELS_PATH, exist_ok=True)

    for url in ['https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/de430.bsp',
                'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0011.tls',
                'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/pck00010.tpc']:
        path = os.path.join(SPICE_KERNELS_PATH, url.split('/')[-1])
        if not os.path.isfile(path):
            download_file(url, path)
        spice.furnsh(path)


def datetime_to_solarday_number(date, startdate, body, longitude, day_cache=None):
    """
    Calculates the number of solar days that have passed since a certain date for an observer on a certain planet.

    :param date: Current date
    :param startdate: Starting date (e.g. date of landing)
    :param body: Solar system body on which the observer is located
    :param longitude: Longitude of the observer location on that planet
    """
    if day_cache is None:
        day_cache = []

    if len(day_cache) == 0 or day_cache[-1] < date:
        interval = spice.Cell_Double(2)
        begindate = startdate if len(day_cache) == 0 else day_cache[-1]
        spice.wninsd(spice.datetime2et(begindate), spice.datetime2et(date + dt.timedelta(days=60)), interval)
        result = spice.gfposc('SUN', 'IAU_{}'.format(body), 'NONE', body, 'LATITUDINAL', 'LONGITUDE',
                              '=', math.radians(longitude + 180), 0, 60, 750, interval)
        for i in range(0, len(result), 2):
            day_cache.append(et2datetime(result[i]))

    return np.searchsorted(np.array(day_cache), date) + 1


def et2datetime(et):
    return dateutil.parser.parse(spice.et2utc(et, 'ISOC', 1) + 'Z').replace(tzinfo=None)


def datetime_to_lst(date, body, longitude):
    h, m, s, _, _ = spice.et2lst(spice.datetime2et(date), spice.bodn2c(body), math.radians(longitude), 'PLANETOCENTRIC')
    return dt.time(h, m, s)


def lst_to_datetime(solarday, lst, startdate, body, longitude, day_cache=None):
    if solarday == 1 and lst < dt.time(0, 0, 1):
        return startdate

    if day_cache is None:
        day_cache = []

    if len(day_cache) < solarday:
        interval = spice.Cell_Double(2)
        begindate = startdate if len(day_cache) == 0 else day_cache[-1]
        spice.wninsd(spice.datetime2et(begindate),
                     spice.datetime2et(begindate + dt.timedelta(days=30) * (solarday - len(day_cache))), interval)
        result = spice.gfposc('SUN', 'IAU_{}'.format(body), 'NONE', body, 'LATITUDINAL', 'LONGITUDE',
                              '=', math.radians(longitude + 180), 0, 60, 750, interval)
        for i in range(0, len(result), 2):
            day_cache.append(dateutil.parser.parse(spice.et2utc(result[i], 'ISOC', 1) + 'Z').replace(tzinfo=None))

    if lst < dt.time(0, 0, 1):
        return day_cache[solarday - 2]

    angle = longitude + 180 - (lst.hour * 3600 + lst.minute * 60 + lst.second) / 86400 * 360

    interval = spice.Cell_Double(2)
    begindate = day_cache[solarday - 2] if solarday > 1 else startdate
    spice.wninsd(spice.datetime2et(begindate), spice.datetime2et(day_cache[solarday - 1]), interval)
    result = spice.gfposc('SUN', 'IAU_{}'.format(body), 'NONE', body, 'LATITUDINAL', 'LONGITUDE',
                          '=', math.radians(angle), 0, 60, 750, interval)
    return et2datetime(result[0])